package eu.jonahbauer.wizard.core.machine;

import eu.jonahbauer.wizard.core.machine.states.State;
import org.jetbrains.annotations.Blocking;
import org.jetbrains.annotations.NotNull;

import java.util.concurrent.*;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

public abstract class Context<T extends State> {
    protected final ScheduledExecutorService scheduler = Executors.newSingleThreadScheduledExecutor();
    protected T state;
    protected final ReentrantLock lock = new ReentrantLock();
    private final Condition finishCondition = lock.newCondition();
    private boolean finished;
    private Throwable exception;

    protected void start(@NotNull T state) {
        lock.lock();
        try {
            if (finished) throw new IllegalStateException("Context has already finished.");
            transition(null, state);
        } finally {
            lock.unlock();
        }
    }

    public void transition(T currentState, T newState) {
        lock.lock();
        try {
            if (state == currentState) {
                state = newState;
                if (currentState != null) currentState.onExit();
                onTransition(currentState, newState);
                if (newState != null) newState.onEnter();
            } else {
                throw new IllegalStateException("Current state does not match.");
            }
        } catch (Throwable t) {
            handleError(t);
        } finally {
            lock.unlock();
        }
    }

    public void finish() {
        finish(null);
    }

    public void finish(Throwable exception) {
        lock.lock();
        try {
            finished = true;
            this.exception = exception;
            finishCondition.signalAll();
            transition(state, null);
            scheduler.shutdown();
        } finally {
            lock.unlock();
        }
    }

    public void cancel() {
        finish(new CancellationException());
    }

    @Blocking
    public void await() throws InterruptedException, ExecutionException, CancellationException {
        lock.lock();
        try {
            while (!finished) {
                finishCondition.await();
            }
            if (exception != null) {
                if (exception instanceof CancellationException cancelled) {
                    throw cancelled;
                } else {
                    throw new ExecutionException(exception);
                }
            }
        } finally {
            lock.unlock();
        }
    }

    public void timeout(@NotNull T currentState, long delay) {
        scheduler.schedule(() -> {
            lock.lock();
            try {
                if (state == currentState) {
                    state.onTimeout();
                }
            } catch (Throwable t) {
                handleError(t);
            } finally {
                lock.unlock();
            }
        }, delay, TimeUnit.MILLISECONDS);
    }

    protected void handleError(Throwable t) {
        finish(t);
    }

    protected void onTransition(T from, T to) {}
}
