package eu.jonahbauer.wizard.core.machine;


import eu.jonahbauer.wizard.core.model.Card;
import eu.jonahbauer.wizard.core.util.Pair;
import lombok.*;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.jetbrains.annotations.Unmodifiable;

import java.util.*;

@With
@Getter
@EqualsAndHashCode
@AllArgsConstructor(access = AccessLevel.PRIVATE)
@SuppressWarnings({"unused", "UnusedReturnValue", "OptionalUsedAsFieldOrParameterType", "OptionalAssignedToNull"})
public final class GameData {
    private final @Unmodifiable List<UUID> players;
    private final @Unmodifiable Map<UUID, Integer> score;

    private final Integer round;
    private final @Unmodifiable Map<UUID, @Unmodifiable List<Card>> hands;
    private final Optional<Card> trumpCard;
    private final Card.Suit trumpSuit;
    private final @Unmodifiable Map<UUID, Integer> predictions;
    private final @Unmodifiable Map<UUID, Integer> tricks;

    private final Integer trick;
    private final @Unmodifiable List<Pair<UUID, Card>> stack;

    private final UUID currentPlayer;

    private @EqualsAndHashCode.Exclude transient boolean requiredPlayers;
    private @EqualsAndHashCode.Exclude transient boolean requiredRound;
    private @EqualsAndHashCode.Exclude transient boolean requiredTrick;
    private @EqualsAndHashCode.Exclude transient boolean requiredHands;
    private @EqualsAndHashCode.Exclude transient boolean requiredTrumpCard;
    private @EqualsAndHashCode.Exclude transient boolean requiredTrumpSuit;
    private @EqualsAndHashCode.Exclude transient boolean requiredStack;
    private @EqualsAndHashCode.Exclude transient boolean requiredPredictions;
    private @EqualsAndHashCode.Exclude transient boolean requiredTricks;
    private @EqualsAndHashCode.Exclude transient boolean requiredScore;
    private @EqualsAndHashCode.Exclude transient boolean requiredCurrentPlayer;

    @Builder(toBuilder = true)
    private GameData(@Unmodifiable List<UUID> players,
                     @Unmodifiable Map<UUID, Integer> score,
                     Integer round,
                     @Unmodifiable Map<UUID, @Unmodifiable List<Card>> hands,
                     Optional<Card> trumpCard,
                     Card.Suit trumpSuit,
                     @Unmodifiable Map<UUID, Integer> predictions,
                     @Unmodifiable Map<UUID, Integer> tricks,
                     Integer trick,
                     @Unmodifiable List<Pair<UUID, Card>> stack,
                     UUID currentPlayer)
    {
        this.players = players != null ? List.copyOf(players) : null;
        this.score = score != null ? Map.copyOf(score) : null;

        this.round = round;
        this.hands = hands != null ? Map.copyOf(hands) : null;
        this.trumpCard = trumpCard;
        this.trumpSuit = trumpSuit;
        this.predictions = predictions != null ? Map.copyOf(predictions) : null;
        this.tricks = tricks != null ? Map.copyOf(tricks) : null;

        this.trick = trick;
        this.stack = stack != null ? List.copyOf(stack) : null;

        this.currentPlayer = currentPlayer;
    }

    //<editor-fold desc="Getters" defaultstate="collapsed">
    public List<UUID> getPlayers() {
        if (players == null) throw new UnsupportedOperationException();
        return players;
    }

    public int getRound() {
        if (round == null) throw new UnsupportedOperationException();
        return round;
    }

    public int getTrick() {
        if (trick == null) throw new UnsupportedOperationException();
        return trick;
    }

    public Map<UUID, List<Card>> getHands() {
        if (hands == null) throw new UnsupportedOperationException();
        return hands;
    }

    public Card getTrumpCard() {
        if (trumpCard == null) throw new UnsupportedOperationException();
        return trumpCard.orElse(null);
    }

    public Card.Suit getTrumpSuit() {
        if (trumpSuit == null) throw new UnsupportedOperationException();
        return trumpSuit;
    }

    public List<Pair<UUID, Card>> getStack() {
        if (stack == null) throw new UnsupportedOperationException();
        return stack;
    }

    public Map<UUID, Integer> getPredictions() {
        if (predictions == null) throw new UnsupportedOperationException();
        return predictions;
    }

    public Map<UUID, Integer> getTricks() {
        if (tricks == null) throw new UnsupportedOperationException();
        return tricks;
    }

    public Map<UUID, Integer> getScore() {
        if (score == null) throw new UnsupportedOperationException();
        return score;
    }

    public UUID getCurrentPlayer() {
        if (currentPlayer == null) throw new UnsupportedOperationException();
        return currentPlayer;
    }

    public UUID getPlayer(int index) {
        return getPlayers().get(index);
    }

    public int getPlayerCount() {
        return getPlayers().size();
    }
    //</editor-fold>

    //<editor-fold desc="Requirements" defaultstate="collapsed">
    public GameData requirePlayers() {
        if (players == null) throw new AssertionError();
        requiredPlayers = true;
        return this;
    }

    public GameData requireRound() {
        if (round == null) throw new AssertionError();
        requiredRound = true;
        return this;
    }

    public GameData requireTrick() {
        if (trick == null) throw new AssertionError();
        requiredTrick = true;
        return this;
    }

    public GameData requireHands() {
        if (hands == null) throw new AssertionError();
        requiredHands = true;
        return this;
    }

    public GameData requireAllHands() {
        requirePlayers();
        requireHands();
        for (UUID uuid : players) {
            if (!hands.containsKey(uuid)) {
                throw new AssertionError();
            }
        }
        return this;
    }

    public GameData requireTrumpCard() {
        if (trumpCard == null) throw new AssertionError();
        requiredTrumpCard = true;
        return this;
    }

    public GameData requireTrumpSuit() {
        if (trumpSuit == null) throw new AssertionError();
        requiredTrumpSuit = true;
        return this;
    }

    public GameData requireStack() {
        if (stack == null) throw new AssertionError();
        requiredStack = true;
        return this;
    }

    public GameData requirePredictions() {
        if (predictions == null) throw new AssertionError();
        requiredPredictions = true;
        return this;
    }

    public GameData requireAllPredictions() {
        requirePlayers();
        requirePredictions();
        for (UUID uuid : players) {
            if (!predictions.containsKey(uuid)) {
                throw new AssertionError();
            }
        }
        return this;
    }

    public GameData requireTricks() {
        if (tricks == null) throw new AssertionError();
        requiredTricks = true;
        return this;
    }

    public GameData requireScore() {
        if (score == null) throw new AssertionError();
        requiredScore = true;
        return this;
    }

    public GameData requireCurrentPlayer() {
        if (currentPlayer == null) throw new AssertionError();
        requiredCurrentPlayer = true;
        return this;
    }

    public GameData onlyRequired() {
        var builder = toBuilder();
        if (!requiredPlayers) builder.players(null);
        if (!requiredRound) builder.round(null);
        if (!requiredTrick) builder.trick(null);
        if (!requiredHands) builder.hands(null);
        if (!requiredTrumpCard) builder.trumpCard(null);
        if (!requiredTrumpSuit) builder.trumpSuit(null);
        if (!requiredStack) builder.stack(null);
        if (!requiredPredictions) builder.predictions(null);
        if (!requiredTricks) builder.tricks(null);
        if (!requiredScore) builder.score(null);
        if (!requiredCurrentPlayer) builder.currentPlayer(null);
        requiredPlayers = false;
        requiredRound = false;
        requiredTrick = false;
        requiredHands = false;
        requiredTrumpCard = false;
        requiredTrumpSuit = false;
        requiredStack = false;
        requiredPredictions = false;
        requiredTricks = false;
        requiredScore = false;
        requiredCurrentPlayer = false;
        return builder.build();
    }
    //</editor-fold>

    //<editor-fold desc="Wither" defaultstate="collapsed">
    public GameData withPrediction(UUID player, int prediction) {
        Map<UUID, Integer> predictions = new HashMap<>(getPredictions());
        predictions.put(player, prediction);
        return withPredictions(predictions);
    }

    public GameData withNextPlayer() {
        int index = getPlayers().indexOf(getCurrentPlayer());
        UUID next;
        if (index == -1) {
            throw new IllegalArgumentException();
        } else if (index == getPlayerCount() - 1) {
            next = getPlayers().get(0);
        } else {
            next = getPlayers().get(index + 1);
        }
        return withCurrentPlayer(next);
    }

    public GameData withNextTrick() {
        int trick = getTrick();
        return withTrick(trick + 1);
    }

    public GameData withNextRound() {
        int round = getRound();
        return withRound(round + 1);
    }

    public GameData withCardPlayed(UUID player, Card card) {
        List<Card> hand = new ArrayList<>(getHands().get(player));
        hand.remove(card);
        Map<UUID, List<Card>> hands = new HashMap<>(getHands());
        hands.put(player, hand);
        List<Pair<UUID, Card>> stack = new ArrayList<>(getStack());
        stack.add(Pair.of(player, card));
        return withStack(stack).withHands(hands);
    }
    //</editor-fold>

    public String toString() {
        StringBuilder builder = new StringBuilder("(");
        if (players != null) builder.append("players=").append(players).append(", ");
        if (round != null) builder.append("round=").append(round).append(", ");
        if (trick != null) builder.append("trick=").append(trick).append(", ");
        if (hands != null) builder.append("hands=").append(hands).append(", ");
        if (trumpCard != null) builder.append("trumpCard=").append(trumpCard).append(", ");
        if (trumpSuit != null) builder.append("trumpSuit=").append(trumpSuit).append(", ");
        if (stack != null) builder.append("stack=").append(stack).append(", ");
        if (predictions != null) builder.append("predictions=").append(predictions).append(", ");
        if (tricks != null) builder.append("tricks=").append(tricks).append(", ");
        if (currentPlayer != null) builder.append("currentPlayer=").append(currentPlayer).append(", ");
        if (builder.length() > 1) builder.setLength(builder.length() - 2);
        builder.append(")");
        return builder.toString();
    }

    // An interface for @lombok.experimental.Delegate
    public interface Getters {
        @NotNull List<UUID> getPlayers();
        @NotNull Map<UUID, Integer> getScore();

        int getRound();
        @NotNull Map<UUID, List<Card>> getHands();
        @Nullable Card getTrumpCard();
        @NotNull Card.Suit getTrumpSuit();
        @NotNull Map<UUID, Integer> getPredictions();
        @NotNull Map<UUID, Integer> getTricks();

        int getTrick();
        @NotNull List<Pair<UUID, Card>> getStack();

        @NotNull UUID getCurrentPlayer();
        @NotNull UUID getPlayer(int index);
        int getPlayerCount();
    }
}
