package eu.jonahbauer.wizard.core.machine.states.trick;

import eu.jonahbauer.wizard.core.machine.Game;
import eu.jonahbauer.wizard.core.machine.GameData;
import eu.jonahbauer.wizard.core.machine.states.round.FinishingRound;
import eu.jonahbauer.wizard.core.messages.observer.TrickMessage;
import eu.jonahbauer.wizard.core.model.Card;
import eu.jonahbauer.wizard.core.util.Pair;
import org.jetbrains.annotations.VisibleForTesting;

import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.UUID;

public final class FinishingTrick extends TrickState {
    private transient UUID winner;

    public FinishingTrick(Game game, GameData data) {
        super(game, data.requireStack());
        assert assertions();
    }

    @Override
    public void onEnter() {
        this.winner = getWinner();

        var tricks = new HashMap<>(getTricks());
        tricks.compute(winner, (k, t) -> t == null ? 1 : t + 1);
        GameData data = getData().withTricks(tricks);

        if (getTrick() < getRound()) {
            transition(new StartingTrick(
                    getGame(),
                    data.withCurrentPlayer(winner)
                        .withNextTrick()
            ));
        } else {
            transition(new FinishingRound(getGame(), data));
        }
    }

    @Override
    public void onExit() {
        if (winner != null) {
            getGame().notify(new TrickMessage(winner, getStack().stream().map(Pair::second).toList()));
        }
    }

    @VisibleForTesting
    UUID getWinner() {
        var wizard = getStack().stream()
                               .filter(pair -> pair.second().getSuit() == Card.Suit.WIZARD)
                               .findFirst()
                               .orElse(null);
        if (wizard != null) {
            return wizard.first();
        }

        if (getStack().stream().allMatch(pair -> pair.second().getSuit() == Card.Suit.JESTER)) {
            return getStack().get(0).first();
        }

        var trumpSuit = getTrumpSuit();
        var suit = getTrickSuit();
        return getStack().stream()
                         .max(
                                 Comparator.<Pair<UUID, Card>>comparingInt(pair -> pair.second()
                                                                                       .getSuit() == trumpSuit ? 1 : 0)
                                           .thenComparing(pair -> pair.second().getSuit() == suit ? 1 : 0)
                                           .thenComparing(pair -> pair.second().getValue())
                         )
                         .orElseThrow(AssertionError::new)
                         .first();
    }

    private boolean assertions() {
        if (getStack().size() != getPlayerCount()) {
            throw new AssertionError("Unexpected stack size: " + getStack().size() + " (expected " + getPlayerCount() + ").");
        }
        int expectedSize = getRound() - getTrick();
        for (List<Card> hand : getHands().values()) {
            if (hand.size() != expectedSize) {
                throw new AssertionError("Unexpected hand size: " + hand.size() + " (expected " + expectedSize + ").");
            }
        }
        return true;
    }
}
